package org.spagic3.client.osgi;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Map;

import javax.activation.DataHandler;
import javax.activation.FileDataSource;
import javax.mail.util.ByteArrayDataSource;

import org.osgi.framework.BundleContext;
import org.osgi.framework.InvalidSyntaxException;
import org.osgi.framework.ServiceReference;
import org.spagic3.client.api.Client;
import org.spagic3.client.api.ClientMessage;
import org.spagic3.client.api.IAttachment;
import org.spagic3.client.api.MessageAttachment;
import org.spagic3.constants.SpagicConstants;
import org.spagic3.core.ExchangeUtils;
import org.spagic3.core.ISpagicService;
import org.spagic3.nmr.api.Exchange;
import org.spagic3.nmr.api.Message;
import org.spagic3.nmr.api.Pattern;

//import com.sun.istack.internal.ByteArrayDataSource;

/**
 * Class <code>OSGiClientImpl</code> an OSGi compliant implementation
 * of the client interface. Some methods throwm 
 *
 * @author <a href="mailto:andrea.zoppello@eng.it">Andrea Zoppello</a>
 * @version $Revision$
 */
public class OSGiClientImpl implements Client {

	private BundleContext context = null;

	/**
	 * Creates a new instance of <code>OSGiClientImpl</code> .
	 *
	 * @param context a <code>BundleContext</code> this bundle's execution
	 * context within the Framework
	 */
	public OSGiClientImpl(BundleContext context){
		this.context = context;
	}

	/**
	 * <code>fireAndForget</code> Fire and forget invocation of a Spagic
	 * service.
	 *
	 * @param spagicServiceId a <code>String</code> the service to be
	 * invoked
	 * @param message a <code>ClientMessage</code> the message to be passed
	 * to the service
	 */
	public void fireAndForget(String spagicServiceId, ClientMessage message) {
			IClientServiceInvoker clientServiceInvoker = OSGiSideClientActivator.getClientServiceInvoker();
			Exchange inOnlyExchange = createASyncInOnlyExchangeFromClientMessage(message);
			clientServiceInvoker.fireAndForget(spagicServiceId, inOnlyExchange);

	}

	/**
	 * <code>invokeAndWait</code> Blocking invocation of a Spagic
	 * service. <b>BEWARE:</b> prepare to catch the RuntimeException
	 * RuntimeException when dealing
	 * with File based attachment: this exception wraps as
	 * RuntimeException any java.io.IOException that may be thrown
	 * during the File based attachment processing.
	 *
	 * @param spagicServiceId a <code>String</code> the service to be
	 * invoked
	 * @param message a <code>ClientMessage</code> the message to be passed
	 * to the service
	 * @return a <code>ClientMessage</code> with the service response.
	 * @exception RuntimeException
	 * (RuntimeException, unchecked) wraps any IOException that may be
	 * thrown while handling the attachments
	 */
	public ClientMessage invokeAndWait(String spagicServiceId,
			ClientMessage message) {

		ServiceReference[] refs = null;
		try {

			refs = context.getServiceReferences(ISpagicService.class.getName(),
					"(" + SpagicConstants.SPAGIC_ID_PROPERTY + "="
							+ spagicServiceId + ")");
		} catch (InvalidSyntaxException e) {
			throw new IllegalStateException(e.getMessage(), e);
		}

		if (refs == null || (refs.length == 0))
			throw new IllegalStateException("Service with spagic.id ["
					+ spagicServiceId + "] not found");
		if (refs != null && (refs.length > 1))
			throw new IllegalStateException(
					"Founded more than one service with the same spagicId ["
							+ spagicServiceId + "]");

		ISpagicService service = null;

		if (refs[0] != null) {
			service = (ISpagicService) context.getService(refs[0]);

			if (service != null) {
				try {
					Exchange inOutExchange = createSyncInOutExchangeFromClientMessage(message);
					service.process(inOutExchange);
					return createClientMessageFromExchange(inOutExchange);
				} finally {
					context.ungetService(refs[0]);
				}
			}else{
				throw new IllegalStateException(
						"Service is null");
			}
		}else{
			throw new IllegalStateException(
			"Service Reference is null");
		}
	}

	/**
	 * <code>fillMessageFromClientMessage</code> creates a message
	 * from the contents of a ClientMessage
	 *
	 * @param msg a <code>Message</code> the resulting Message
	 * @param clientMessage a <code>ClientMessage</code> the source
	 * ClientMessage 
	 */
	public void fillMessageFromClientMessage(Message msg, ClientMessage clientMessage){
		msg.setBody(clientMessage.getBody());

		Map<String, String> clientMessageProperties = clientMessage.getProperties();
		if (clientMessageProperties != null){
			for (String key : clientMessageProperties.keySet()){
				msg.setHeader(key, clientMessageProperties.get(key));
			}
		}
		
		Map<String, IAttachment> clientMessageAttachments = clientMessage.getAttachments();
		if (clientMessageAttachments != null){
			for (String key : clientMessageAttachments.keySet()){
				msg.addAttachment(key, toDataHandler(clientMessageAttachments.get(key),key));
			}
		}
	}


	/**
	 * <code>fillClientMessageFromMessage</code> fills a ClientMessage
	 * from the contents of a Message. <b>BEWARE:</b> prepare to catch
	 * the RuntimeException when dealing
	 * with File based attachment: this exception wraps as
	 * RuntimeException any java.io.IOException that may be thrown
	 * during the File based attachment processing.
	 *
	 * @param clientMessage a <code>ClientMessage</code> the
	 * ClientMessage to be filled.
	 * @param from a <code>Message</code> the source Message
	 * @return a <code>ClientMessage</code> the ClientMessage filled.
	 * @exception RuntimeException
	 * (RuntimeException, unchecked) wraps any IOException that may be
	 * thrown while handling the attachments
	 */
	public ClientMessage fillClientMessageFromMessage(ClientMessage clientMessage, Message from ) {

		clientMessage.setBody((String)from.getBody());

		Map<String, Object> clientMessageProperties = from.getHeaders();
		for (String key : clientMessageProperties.keySet()){
			clientMessage.setProperty(key, from.getHeader(key).toString());
		}

		Map<String, Object> outAtt = from.getAttachments();

		Object att = null;
		for (String key : outAtt.keySet()){
			att = outAtt.get(key);
			if ( att instanceof DataHandler){
				try {
					clientMessage.setAttachment(key, fromDataHandler((DataHandler)att));
				}
				catch (IOException jiioex) {
					throw new RuntimeException("Failure in managing a file based attachment", jiioex);
				}
			}
		}
		return clientMessage;
	}

	/**
	 * <code>createSyncInOutExchangeFromClientMessage</code> creates a
	 * synchronous Exchange with the contents of a ClientMessage
	 *
	 * @param clientMessage a <code>ClientMessage</code> the source
	 * ClientMessage
	 * @return an <code>Exchange</code> a new Exchange created from the
	 * ClientMessage
	 */
	public Exchange createSyncInOutExchangeFromClientMessage(ClientMessage clientMessage){
		Exchange exchange= ExchangeUtils.createExchange(Pattern.InOut);

		exchange.setProperty(SpagicConstants.SYNC_EXCHANGE, "true");
		Message in = exchange.getIn(true);
		fillMessageFromClientMessage(in, clientMessage);
		exchange.setIn(in);

		return exchange;
	}

	/**
	 * <code>createASyncInOnlyExchangeFromClientMessage</code> creates an
	 * asynchronous Exchange with the contents of a ClientMessage
	 *
	 * @param clientMessage a <code>ClientMessage</code> the source
	 * ClientMessage
	 * @return an <code>Exchange</code> a new Exchange created from the
	 * ClientMessage
	 */
	public Exchange createASyncInOnlyExchangeFromClientMessage(ClientMessage clientMessage){
		Exchange exchange= ExchangeUtils.createExchange(Pattern.InOnly);


		Message in = exchange.getIn(true);
		fillMessageFromClientMessage(in, clientMessage);
		exchange.setIn(in);

		return exchange;
	}

	/**
	 * <code>createClientMessageFromExchange</code> creates a
	 * ClientMessage from the contents of an Exchange. <b>BEWARE:</b>
	 * prepare to catch the RuntimeException
	 * RuntimeException when dealing
	 * with File based attachment: this exception wraps as
	 * RuntimeException any java.io.IOException that may be thrown
	 * during the File based attachment processing.
	 *
	 * @param exchange an <code>Exchange</code> the source Exchange
	 * @return a <code>ClientMessage</code> a new ClientMessage created
	 * from the Exchange
	 * @exception RuntimeException
	 * (RuntimeException, unchecked) wraps any IOException that may be
	 * thrown while handling the attachments
	 */
	public ClientMessage createClientMessageFromExchange(Exchange exchange) {

		Message out = exchange.getOut(false);

		if (out == null)
			throw new IllegalStateException("Null Response From Service");

		ClientMessage response = new ClientMessage(exchange.getId());
		return fillClientMessageFromMessage(response, out);
	}

	private DataHandler toDataHandler (IAttachment that, String name) {
		return new DataHandler(
				(that.isDataAFile()) 
				? new FileDataSource(that.getDataAsFile()) 
				: new ByteArrayDataSource(that.getDataAsByteArray(), name/*, that.getMimeType()*/));
	}
	
	public static IAttachment fromDataHandler(DataHandler from) throws IOException {
		InputStream fromIn = from.getInputStream();
		ByteArrayOutputStream baos = new ByteArrayOutputStream(fromIn.available());
		byte buf[] = new byte[1024];
		int readed = 0;
		
		while( (readed = fromIn.read(buf, 0 , 1024)) > -1 )
			baos.write(buf,0,readed);
		
		IAttachment retVal = new MessageAttachment();
		retVal.setData(baos.toByteArray(), from.getContentType());
		return retVal;
	}

}
